// ====================================================
//
// Copyright (c) 2001 Sean Wilson. All Rights Reserved.
//
// ====================================================

import java.util.Vector;

/** Represents a bat. */
public class Bat
{
	/** Position of bat. */
	private Vector3d position;
	/** Velocity of bat. */
	private Vector3d velocity = new Vector3d();

	/** Mass of bat. */
	private float mass;
	/** Maximum steering force bat can excert. */
	private float maxForce;
	/** Current steering direction. */
	private Vector3d steering = new Vector3d();

	/** Current wing flapping stage. */
	private float flapStage = (float)Math.random();
	/** Minimum change of flap stage per second. */
	private float minFlapPerSecond;
	/** Change of flap stage per unit velocity. */
	private float flapPerVelocity;

	/** Constructs a bat.
	 *  @parm initPosition         initial position.
	 *  @parm initMass             mass.
	 *  @parm initMaxForce         maximum steering force.
	 *  @parm initMinFlapPerSecond minimum change of flap stage per second.
	 *  @parm initFlapPerVelocity  change of flap stage per unit velocity.
	 */
	public Bat(Vector3d initPosition, float initMass, float initMaxForce, float initMinFlapPerSecond, float initFlapPerVelocity)
	{
		position         = new Vector3d(initPosition);
		mass             = initMass;
		maxForce         = initMaxForce;
		minFlapPerSecond = initMinFlapPerSecond;
		flapPerVelocity  = initFlapPerVelocity;
	}

	/** Adds steering force.
	 *  @param weight    weight of steering force relative to others added.
	 *  @param direction direction of steering force.
	 */
	public void addSteering(float weight, Vector3d direction)
	{
		// Add steering in given direction with same length as steering weight
		steering = steering.add(direction.resize(weight));
	}

	/** Adds flocking steering. If no neighbours are found to flock with, join steering is used to seek nearest bat.
	 *  @param batList         list of bats that can be flock with.
	 *  @param neighbourRadius distance bat can be from this bat to be considered a neighbour.
	 *  @param cohereWeight    weight of cohersion steering with neighbours.
	 *  @param seperateWeight  weight of seperation steering with neighbours.
	 *  @param alignWeight     weight of alignment steering with neighbours.
	 *  @param joinWeight      weight of join nearest bat steering.
	 */
	public void flock(Vector batList, float neighbourRadius, float cohereWeight, float seperateWeight, float alignWeight, float joinWeight)
	{
		// Whether neighbours were found
		boolean hasNeighbour = false;
		// Cohersion steering
		Vector3d cohereSteer = new Vector3d();
		// Seperation steering
		Vector3d seperateSteer = new Vector3d();
		// Alignment steering
		Vector3d alignSteer = new Vector3d();

		// Index of closest bat
		int closestBat = 0;
		// Distance of closest bat
		float closestDistance = Integer.MIN_VALUE;
		
		// For each bat
		for (int i = 0; i < batList.size(); ++i)
		{
			// Get current bat
			Bat bat = (Bat)batList.elementAt(i);

			// Goto next bat if current bat is this bat
			if (bat == this)
				continue;

			// Get vector from this bat to current bat
			Vector3d offset = new Vector3d(position, bat.getPosition());
			// Calculate distance between the bats
			float distance = offset.getLength();
			
			// If bat is a neigbour
			if (distance <= neighbourRadius)
			{
				// Neighbour was found
				hasNeighbour = true;

				// Add flocking steering
				cohereSteer   = cohereSteer.add(offset);
				seperateSteer = seperateSteer.subtract(offset);
				alignSteer    = alignSteer.add(bat.getVelocity());
			}
			else
			{
				// If current bat is closest than previous closest
				if (distance < closestDistance)
				{
					// Current bat is new closest bat
					closestBat      = i;
					closestDistance = distance;
				}
			}
		}

		// If bat has neighbours
		if (hasNeighbour)
		{
			// Add flocking steering
			addSteering(cohereWeight, cohereSteer);
			addSteering(seperateWeight, seperateSteer);
			addSteering(alignWeight, alignSteer);
		}
		else
		{
			// Add steering to join closest bat
			addSteering(joinWeight, new Vector3d(position, ((Bat)batList.elementAt(closestBat)).getPosition()));
		}
	}

	/** Adds steering to cancel vertical velocity componenet.
	 *  @param weight weight of steering.
	 */
	public void cancelVerticalVelocity(float weight)
	{
		// Steer to cancel vertical velocity
		addSteering(weight, new Vector3d(0, (velocity.y > 0) ? -1 : +1, 0));
	}

	/** Adds steering towards center of an enclosing sphere if bat ventures outside it.
	 *  @param weight weight of steering.
	 *  @param center center of enclosing sphere.
	 *  @param radius radius of enclosing sphere.
	 */
	public void enclose(float weight, Vector3d center, float radius)
	{
		// Get vector from position to center of enclosing sphere
		Vector3d offset = new Vector3d(position, center);
		
		// If outside sphere
		if (offset.getLength() > radius)
		{
			// Steer towards sphere center
			addSteering(weight, offset);
		}
	}

	/** Updates bat's position and resets steering force.
	 *  @param deltaTime time step to update by.
	 *  @param damping   damping coefficient.
	 */
	public void update(float deltaTime, float damping)
	{
		// Calculate net force by combining steering force and damping force
		Vector3d netForce = steering.normalise().multiply(maxForce).add(velocity.multiply(-damping));
		// Calculate acceleration caused by force
		Vector3d acceleration = netForce.divide(mass);

		// Calculate new velocity
		velocity = velocity.add(acceleration.multiply(deltaTime));
		// Calculate new position
		position = position.add(velocity.multiply(deltaTime));
		// Reset steering direction
		steering = new Vector3d();

		// Update flap stage
		flapStage += ((minFlapPerSecond/* + velocity.getLength() * flapPerVelocity*/) * deltaTime) % 1;
	}

	/** Returns bat's position. */
	Vector3d getPosition()
	{
		return new Vector3d(position);
	}

	/** Returns bat's velocity. */
	Vector3d getVelocity()
	{
		return new Vector3d(velocity);
	}

	/** Returns bat's flap stage, where 0 is the start of a flap and 1 is the end. */
	float getFlapStage()
	{
		return flapStage;
	}
}