// ====================================================
//
// Copyright (c) 2001 Sean Wilson. All Rights Reserved.
//
// ====================================================

import java.awt.*;
import java.util.*;
import java.applet.*;

public class BatsApplet extends Applet implements Runnable
{
	/** Radius that neighbours are contained in. */
	private final int NEIGHBOUR_RADIUS = 5;	
		
	/** Weighting of seperation steering. */
	private final float SEPERATE_WEIGHT = 10;
	/** Weighting of alignment steering. */
	private final float ALIGN_WEIGHT = 6;
	/** Weighting of cohersion steering. */
	private final float COHERE_WEIGHT = 5;
	/** Weighting of joining steering. */
	private final float JOIN_WEIGHT = 10;
	/** Weighting of vertical velocity stopping steering. */
	private final float VERTICAL_WEIGHT = 1;
	/** Weighting of enclosing steering. */
	private final float ENCLOSE_WEIGHT = 10;
	/** Center of enclosing sphere. */
	private final Vector3d ENCLOSE_CENTER = new Vector3d(0, 0, 50);
	/** Enclosing sphere radius. */
	private final float ENCLOSE_RADIUS = 20;

	/** Point bats enter world. */
	private final Vector3d ENTRY_POINT = new Vector3d(25, 25, 50);

	/** Damping coefficient. */
	private final float DAMPING = 0.3f;
	/** Time step to take each update. */
	private final float TIME_STEP = 0.05f;

	/** Number of bats in world. */
	private final float NUM_BATS = 64;
	/** Mass of each bat. */
	private final float BAT_MASS = 1;
	/** Maximum force bat can excert. */
	private final float BAT_FORCE = 3;

	/** Minimum bat flap stage change per second. */
	private final float MIN_FLAP_PER_SECOND = 0.8f;
	/** Bat flap stage change per second. */
	private final float FLAP_PER_VELOCITY = 1.0f;

	/** Filename of background image. */
	private final String BACKGROUND_IMAGE_FILENAME = "background.jpg";
	
	/** Thread used for updating application. */
	private Thread updateThread;
	
	/** Double buffer image. */
	private Image offscreen;
	/** Background image. */
	private Image background;

	/** List of bats in world. */
	private Vector batList = new Vector();

	/** List of vertices of a bat wing along x-axis facing positive direction. */
	Vector3d[] wing = {new Vector3d(+0.29f, 0, +0.00f),
					   new Vector3d(+0.24f, 0, +0.02f),
					   new Vector3d(+0.31f, 0, +0.27f),
					   new Vector3d(+0.42f, 0, +0.35f),
					   new Vector3d(+0.32f, 0, +0.60f),
					   new Vector3d(+0.20f, 0, +0.75f),
					   new Vector3d(-0.07f, 0, +1.00f),
					   new Vector3d(-0.05f, 0, +0.90f),
					   new Vector3d(-0.09f, 0, +0.82f),
					   new Vector3d(-0.19f, 0, +0.75f),
					   new Vector3d(-0.18f, 0, +0.65f),
					   new Vector3d(-0.20f, 0, +0.53f),
					   new Vector3d(-0.32f, 0, +0.43f),
					   new Vector3d(-0.26f, 0, +0.32f),
					   new Vector3d(-0.26f, 0, +0.23f),
					   new Vector3d(-0.34f, 0, +0.12f),
					   new Vector3d(-0.30f, 0, +0.07f),
					   new Vector3d(-0.35f, 0, +0.03f),
					   new Vector3d(-0.42f, 0, +0.00f)};
		
	/** Width of projection plane. */
	private float planeWidth;
	/** Height of projection plane. */
	private float planeHeight;

	/** Field of Vision angle size. */
	private final float FOV = (float)(50 * (Math.PI / 180));

	/** Whether background image has finished loaded. */
	private boolean backgroundLoaded = false;
	
	public void init()
	{
		// Calculate projection plane dimensions
		planeWidth  = (float)getSize().width  / (2 * (float)Math.tan(FOV * 0.5f));
		planeHeight = (float)getSize().height / (2 * (float)Math.tan(FOV * 0.5f));
		
		// Create offscreen image
		offscreen = createImage(getSize().width, getSize().height);

		// For each bat
		for (int i = 0; i < NUM_BATS; ++i)
		{
			// Bats will enter near entry point in a slightly random horizontal line.
			Vector3d position = ENTRY_POINT.add(new Vector3d(i, (float)Math.random(), (float)Math.random()));
			// Create new bat and add it to list
			batList.addElement(new Bat(position, BAT_MASS, BAT_FORCE, MIN_FLAP_PER_SECOND, FLAP_PER_VELOCITY));
		}
	}

	public void start()
	{
		// Create update thread
		updateThread = new Thread(this);
		// Start update thread
		updateThread.start();
	}

	public void stop()
	{
		// Stop update thread
		updateThread = null;
	}

	public void destroy()
	{
		// Stop update thread
		updateThread = null;
	}

	public void loadBackgroundImage()
	{
		// Set to border layout
		setLayout(new BorderLayout());
		
		// Add wait message label to center of applet
		Label messageLabel = new Label("Loading image...", Label.CENTER);
		add(messageLabel, BorderLayout.CENTER);

		// Show message
		repaint();

		// Background image ID for media tracker
		final int BACKGROUND_IMAGE_ID = 0;
		// Get background image
		background = getImage(getDocumentBase(), BACKGROUND_IMAGE_FILENAME);					

		// Media tracker to wait for image loading
		MediaTracker mediaTracker = new MediaTracker(this);
		// Add image to media tracker
		mediaTracker.addImage(background, BACKGROUND_IMAGE_ID);
		
		try
		{
			// Wait for background image to load
			mediaTracker.waitForID(BACKGROUND_IMAGE_ID);
		}
		catch (InterruptedException e)
		{
		}

		// If error loading background image
		if (mediaTracker.isErrorID(BACKGROUND_IMAGE_ID))
		{
			throw new RuntimeException("Error trying to load " + BACKGROUND_IMAGE_FILENAME);
		}

		// Remove loading message
		remove(messageLabel);

		// Background image has been loaded
		backgroundLoaded = true;
	}

	public void run()
	{
		// Load background image
		loadBackgroundImage();
		
		while (Thread.currentThread() == updateThread)
		{
			// For each bat
			for (int i = 0; i < batList.size(); ++i)
			{
				// Get current bat
				Bat bat = (Bat)batList.elementAt(i);
			
				// Set bat flocking steering
				bat.flock(batList, NEIGHBOUR_RADIUS, COHERE_WEIGHT, SEPERATE_WEIGHT, ALIGN_WEIGHT, JOIN_WEIGHT);
				// Set bat enclosing steering
				bat.enclose(ENCLOSE_WEIGHT, ENCLOSE_CENTER, ENCLOSE_RADIUS);
				// Set bat steering to stop vertical movement
				bat.cancelVerticalVelocity(VERTICAL_WEIGHT);
			}

			// For each bat
			for (int i = 0; i < batList.size(); ++i)
			{
				// Update bat
				((Bat)batList.elementAt(i)).update(TIME_STEP, DAMPING);
			}

			// Repaint scene
			repaint();

			try
			{
				Thread.sleep(10);
			}
			catch (InterruptedException e)
			{
			}
		}
	}

	public void paintScene(Graphics g)
	{
		// Paint background image
		g.drawImage(background, 0, 0, this);
		
		// For each bat
		for (int i = 0; i < batList.size(); ++i)
		{
			// Get current bat
			Bat bat = (Bat)batList.elementAt(i);
			// Calculate bat orientation
			Vector3d orientation = bat.getVelocity().normalise();
			
			// Calculate how to rotate wing to align with bat orientation
			Vector3d batRotate = new Vector3d((float)Math.atan2(orientation.y, orientation.z), (float)Math.atan2(orientation.y, orientation.x), 0);
			// Calculate transformation for each wing vertex, not including wing flapping
			Transform3d worldTransform = new Transform3d().rotate(batRotate).translate(bat.getPosition());

			// Calculate amount wing is to rotate from current flap stage
			float flapRotate = (float)(Math.sin(bat.getFlapStage() * Math.PI * 2) * (Math.PI / 3));

			// Calculate rotation for both wings
			Vector3d wingRotate1 = new Vector3d(+flapRotate,                  0, 0);
			Vector3d wingRotate2 = new Vector3d(-flapRotate + (float)Math.PI, 0, 0);

			// Calculate complete transformations for each wing
			Transform3d wingTransform1 = new Transform3d().rotate(wingRotate1).add(worldTransform);
			Transform3d wingTransform2 = new Transform3d().rotate(wingRotate2).add(worldTransform);

			// Screen co-ord lists for both wings
			int[] xList1 = new int[wing.length];
			int[] yList1 = new int[wing.length];
			int[] xList2 = new int[wing.length];
			int[] yList2 = new int[wing.length];

			// For each wing vertex
			for (int j = 0; j < wing.length; ++j)
			{
				// Calculate both wing vertices in world space
				Vector3d wing1 = wingTransform1.transform(wing[j]);
				Vector3d wing2 = wingTransform2.transform(wing[j]);
								
				// Calculate both wing screen co-ords
				xList1[j] = (int)((wing1.x * planeWidth)  / wing1.z + getSize().width / 2);
				yList1[j] = (int)((wing1.y * planeHeight) / wing1.z + getSize().width / 2);
				xList2[j] = (int)((wing2.x * planeWidth)  / wing2.z + getSize().width / 2);
				yList2[j] = (int)((wing2.y * planeHeight) / wing2.z + getSize().width / 2);
			}

			// Fill both wings polygons black
			g.setColor(Color.black);
			g.fillPolygon(xList1, yList1, xList1.length);
			g.fillPolygon(xList2, yList2, xList2.length);
		}
	}

	public void update(Graphics g)
	{
		// If background image has loaded
		if (backgroundLoaded)
		{
			// Paint double buffer
			paint(g);
		}
		else
		{
			// Paint any components
			paintAll(g);
		}
	}

	public void paint(Graphics g)
	{
		// Paint scene to double buffer
		paintScene(offscreen.getGraphics());
		// Paint double buffer
		g.drawImage(offscreen, 0, 0, this);
	}

	public String getAppletInfo()
	{
		return "Bats - Copyright (c) 2001 Sean Wilson. All Rights Reserved.";
	}
}